@@ -32,10 +32,12 @@
 
 #include <ctype.h>
 
+#include <freetds/utils.h>
 #include <freetds/odbc.h>
 #include <freetds/convert.h>
 #include <freetds/iconv.h>
-#include <freetds/string.h>
+#include <freetds/utils/string.h>
+#include <freetds/encodings.h>
 #include <odbcss.h>
 
 #define TDS_ISSPACE(c) isspace((unsigned char) (c))
@@ -72,18 +74,18 @@ odbc_convert_char(TDS_STMT * stmt, TDSCOLUMN * curcol, TDS_CHAR * src, TDS_UINT
 	if (!conv)
 		conv = tds->conn->char_convs[client2server_chardata];
 	if (desttype == SQL_C_WCHAR) {
-		const char *charset = odbc_get_wide_name(tds->conn);
+		int charset = odbc_get_wide_canonic(tds->conn);
 		/* SQL_C_WCHAR, convert to wide encode */
-		conv = tds_iconv_get(tds->conn, charset, conv->to.charset.name);
+		conv = tds_iconv_get_info(tds->conn, charset, conv->to.charset.canonic);
 		if (!conv)
-			conv = tds_iconv_get(tds->conn, charset, "ISO-8859-1");
+			conv = tds_iconv_get_info(tds->conn, charset, TDS_CHARSET_ISO_8859_1);
 #ifdef ENABLE_ODBC_WIDE
 	} else {
-		conv = tds_iconv_get(tds->conn, tds_dstr_cstr(&stmt->dbc->original_charset), conv->to.charset.name);
+		conv = tds_iconv_get_info(tds->conn, stmt->dbc->original_charset_num, conv->to.charset.canonic);
 		if (!conv)
-			conv = tds_iconv_get(tds->conn, tds_dstr_cstr(&stmt->dbc->original_charset), "ISO-8859-1");
+			conv = tds_iconv_get_info(tds->conn, stmt->dbc->original_charset_num, TDS_CHARSET_ISO_8859_1);
 		if (!conv)
-			conv = tds_iconv_get(tds->conn, "ISO-8859-1", "ISO-8859-1");
+			conv = tds_iconv_get_info(tds->conn, TDS_CHARSET_ISO_8859_1, TDS_CHARSET_ISO_8859_1);
 #endif
 	}
 
@@ -135,7 +137,7 @@ odbc_convert_char(TDS_STMT * stmt, TDSCOLUMN * curcol, TDS_CHAR * src, TDS_UINT
 }
 
 /**
- * Handle conversions from TDS NCHAR to ISO8859-1 striping spaces (for fixed types)
+ * Handle conversions from TDS NCHAR to ISO8859-1 stripping spaces (for fixed types)
  */
 static int
 odbc_tds_convert_wide_iso(TDSCOLUMN *curcol, TDS_CHAR *src, TDS_UINT srclen, TDS_CHAR *buf, TDS_UINT buf_len)
@@ -175,7 +177,7 @@ odbc_tds_convert_wide_iso(TDSCOLUMN *curcol, TDS_CHAR *src, TDS_UINT srclen, TDS
 	return p - buf;
 }
 
-/* The following structure is going to write in these structure not using them
+/* The following function is going to write in these structure not using them
  * but just knowing the ABI. Check these ABI. Mainly make sure the alignment
  * is still correct.
  */
@@ -237,7 +239,7 @@ odbc_convert_datetime_to_binary(TDS_STMT * stmt, TDSCOLUMN *curcol, int srctype,
 	if (destlen == 0)
 		return len;
 
-	cplen = (destlen > len) ? len : destlen;
+	cplen = ODBC_MIN(destlen, len);
 	memcpy(dest, buf, cplen);
 	if (curcol)
 		curcol->column_text_sqlgetdatapos += cplen;
@@ -264,7 +266,7 @@ odbc_convert_to_binary(TDS_STMT * stmt, TDSCOLUMN *curcol, int srctype, TDS_CHAR
 
 	/* if destlen == 0 we return only length */
 	if (destlen > 0) {
-		size_t cplen = (destlen > srclen) ? srclen : destlen;
+		size_t cplen = ODBC_MIN(destlen, srclen);
 		/* do not NUL terminate binary buffer */
 		memcpy(dest, src, cplen);
 		if (curcol)
@@ -273,7 +275,7 @@ odbc_convert_to_binary(TDS_STMT * stmt, TDSCOLUMN *curcol, int srctype, TDS_CHAR
 	return ret;
 }
 
-SQLLEN
+static SQLLEN
 odbc_tds2sql(TDS_STMT * stmt, TDSCOLUMN *curcol, int srctype, TDS_CHAR * src, TDS_UINT srclen, int desttype, TDS_CHAR * dest, SQLULEN destlen,
 	     const struct _drecord *drec_ixd)
 {
@@ -292,22 +294,6 @@ odbc_tds2sql(TDS_STMT * stmt, TDSCOLUMN *curcol, int srctype, TDS_CHAR * src, TD
 
 	assert(desttype != SQL_C_DEFAULT);
 
-	if (curcol) {
-		if (is_blob_col(curcol)) {
-			if (srctype == SYBLONGBINARY && (
-			    curcol->column_usertype == USER_UNICHAR_TYPE ||
-			    curcol->column_usertype == USER_UNIVARCHAR_TYPE))
-				srctype = SYBTEXT;
-			if (curcol->column_type == SYBVARIANT)
-				srctype = ((TDSVARIANT *) src)->type;
-			src = ((TDSBLOB *) src)->textvalue;
-		}
-		if (is_variable_type(curcol->column_type)) {
-			src += curcol->column_text_sqlgetdatapos;
-			srclen -= curcol->column_text_sqlgetdatapos;
-		}
-	}
-
 	nDestSybType = odbc_c_to_server_type(desttype);
 	if (!nDestSybType) {
 		odbc_errs_add(&stmt->errs, "HY003", NULL);
@@ -360,13 +346,7 @@ odbc_tds2sql(TDS_STMT * stmt, TDSCOLUMN *curcol, int srctype, TDS_CHAR * src, TD
 	if (desttype == SQL_C_WCHAR)
 		destlen /= sizeof(SQLWCHAR);
 	if (desttype == SQL_C_CHAR || desttype == SQL_C_WCHAR) {
-		switch (srctype) {
-		case SYBLONGBINARY:
-		case SYBBINARY:
-		case SYBVARBINARY:
-		case SYBIMAGE:
-		case XSYBBINARY:
-		case XSYBVARBINARY:
+		if (is_binary_type(srctype)) {
 			binary_conversion = 1;
 			if (destlen && !(destlen % 2))
 				--destlen;
@@ -433,7 +413,7 @@ odbc_tds2sql(TDS_STMT * stmt, TDSCOLUMN *curcol, int srctype, TDS_CHAR * src, TD
 		}
 
 		nRetVal = strlen(buf);
-		memcpy(dest, buf, destlen < nRetVal ? destlen : nRetVal);
+		memcpy(dest, buf, ODBC_MIN(destlen, nRetVal));
 	} else {
 normal_conversion:
 		nRetVal = tds_convert(context, srctype, src, srclen, nDestSybType, &ores);
@@ -451,7 +431,7 @@ normal_conversion:
 		ret = nRetVal;
 		/* TODO handle not terminated configuration */
 		if (destlen > 0) {
-			cplen = (destlen - 1) > nRetVal ? nRetVal : (destlen - 1);
+			cplen = ODBC_MIN(destlen - 1, nRetVal);
 			assert(cplen >= 0);
 			/*
 			 * odbc always terminate but do not overwrite 
@@ -475,7 +455,7 @@ normal_conversion:
 			SQLWCHAR *wp = (SQLWCHAR *) dest;
 			SQLCHAR  *p  = (SQLCHAR *)  dest;
 
-			cplen = (destlen - 1) > nRetVal ? nRetVal : (destlen - 1);
+			cplen = ODBC_MIN(destlen - 1, nRetVal);
 			assert(cplen >= 0);
 			/*
 			 * odbc always terminate but do not overwrite 
@@ -633,3 +613,32 @@ normal_conversion:
 
 	return ret;
 }
+
+SQLLEN odbc_tds2sql_col(TDS_STMT * stmt, TDSCOLUMN *curcol, int desttype, TDS_CHAR * dest, SQLULEN destlen,
+			const struct _drecord *drec_ixd)
+{
+	int srctype = tds_get_conversion_type(curcol->on_server.column_type, curcol->on_server.column_size);
+	TDS_CHAR *src = (TDS_CHAR *) curcol->column_data;
+	TDS_UINT srclen = curcol->column_cur_size;
+
+	if (is_blob_col(curcol)) {
+		if (srctype == SYBLONGBINARY && (
+		    curcol->column_usertype == USER_UNICHAR_TYPE ||
+		    curcol->column_usertype == USER_UNIVARCHAR_TYPE))
+			srctype = SYBNTEXT;
+		if (srctype == SYBVARIANT)
+			srctype = ((TDSVARIANT *) src)->type;
+		src = ((TDSBLOB *) src)->textvalue;
+	}
+	if (is_variable_type(srctype)) {
+		src += curcol->column_text_sqlgetdatapos;
+		srclen -= curcol->column_text_sqlgetdatapos;
+	}
+	return odbc_tds2sql(stmt, curcol, srctype, src, srclen, desttype, dest, destlen, drec_ixd);
+}
+
+SQLLEN odbc_tds2sql_int4(TDS_STMT * stmt, TDS_INT *src, int desttype, TDS_CHAR * dest, SQLULEN destlen)
+{
+	return odbc_tds2sql(stmt, NULL, SYBINT4, (TDS_CHAR *) src, sizeof(*src),
+			    desttype, dest, destlen, NULL);
+}
