@@ -47,7 +47,7 @@
 #include <freetds/tds.h>
 #include <freetds/convert.h>
 #include <freetds/bytes.h>
-#include "replacements.h"
+#include <freetds/replacements.h>
 
 typedef unsigned short utf16_t;
 
@@ -78,26 +78,43 @@ static const char *parse_numeric(const char *buf, const char *pend,
 
 #define test_alloc(x) {if ((x)==NULL) return TDS_CONVERT_NOMEM;}
 
-#define IS_TINYINT(x) ( 0 <= (x) && (x) <= 0xff )
-#define IS_SMALLINT(x) ( -32768 <= (x) && (x) <= 32767 )
-#define IS_USMALLINT(x) ( 0 <= (x) && (x) <= 65535 )
 /*
+ * Macros for number checks.
+ * They work for signed/unsigned integers and floating point values
+ * (accounting for truncation of any fractional portions of the latter).
+ */
+#define IS_TINYINT(x) ( -1 < (x) && (x) < 0x100 )
+#define IS_SMALLINT(x) ( -32769 < (x) && (x) < 32768 )
+#define IS_USMALLINT(x) ( -1 < (x) && (x) < 65536 )
+
+/*
+ * Macros for integer number checks.
+ * IS_UINT works for both integers and floating point values.
+ *
  * f77: I don't write -2147483648, some compiler seem to have some problem 
  * with this constant although is a valid 32bit value
  */
-#define TDS_INT_MIN (-2147483647l-1l)
-#define TDS_INT_MAX 2147483647l
-#define IS_INT(x) (TDS_INT_MIN <= (x) && (x) <= TDS_INT_MAX)
-
-#define TDS_UINT_MAX 4294967295lu
-#define IS_UINT(x) (0 <= (x) && (x) <= TDS_UINT_MAX)
-
-#define TDS_INT8_MAX ((((TDS_INT8) 0x7fffffffl) << 32) + (TDS_INT8) 0xfffffffflu)
+#define TDS_INT_MIN (-2147483647-1)
+#define TDS_INT_MAX 2147483647
+#define INT_IS_INT(x) (TDS_INT_MIN <= (x) && (x) <= TDS_INT_MAX)
+#define TDS_UINT_MAX 4294967295u
+#define IS_UINT(x) (-1 < (TDS_INT8)(x) && (x) < (TDS_INT8) TDS_UINT_MAX + 1)
 #define TDS_INT8_MIN (-(((TDS_INT8)1)<<62) -(((TDS_INT8)1)<<62))
-#define IS_INT8(x) (TDS_INT8_MIN <= (x) && (x) <= TDS_INT8_MAX)
+#define TDS_INT8_MAX ((((TDS_INT8) 0x7fffffff) << 32) + (TDS_INT8) 0xffffffffu)
 
-#define TDS_UINT8_MAX ((((TDS_UINT8) 0xfffffffflu) << 32) + 0xfffffffflu)
-#define IS_UINT8(x) (0 <= (x) && (x) <= TDS_UINT8_MAX)
+/*
+ * Macros for floating point number checks.
+ * To avoid possible loss of precision, use upper bounds for floating point
+ * values and test MIN - (x) < 1 rather than MIN - 1 < (x) for signed
+ * destination types.
+ */
+#define TDS_INT_UPPER_FLOAT 2147483648.0f
+#define FLOAT_IS_INT(x) (TDS_INT_MIN - (x) < 1.0f && (x) < TDS_INT_UPPER_FLOAT)
+#define TDS_INT8_MIN_FLOAT (-9223372036854775808.0f)
+#define TDS_INT8_UPPER_FLOAT 9223372036854775808.0f
+#define FLOAT_IS_INT8(x) (TDS_INT8_MIN_FLOAT - (x) < 1.0f && (x) < TDS_INT8_UPPER_FLOAT)
+#define TDS_UINT8_UPPER 18446744073709551616.0f
+#define FLOAT_IS_UINT8(x) (-1.0f < (x) && (x) < TDS_UINT8_UPPER)
 
 #define TDS_ISDIGIT(c) ((c) >= '0' && (c) <= '9')
 
@@ -446,7 +463,7 @@ tds_convert_char(const TDS_CHAR * src, TDS_UINT srclen, int desttype, CONV_RESUL
 			cr->m.mny = mymoney;
 			return sizeof(TDS_MONEY);
 		} else {
-			if (!IS_INT(mymoney))
+			if (!INT_IS_INT(mymoney))
 				return TDS_CONVERT_OVERFLOW;
 			cr->m4.mny4 = (TDS_INT) mymoney;
 			return sizeof(TDS_MONEY4);
@@ -713,7 +730,7 @@ tds_convert_int8(const TDS_INT8 *src, int desttype, CONV_RESULT * cr)
 	TDS_CHAR tmp_str[24];
 
 	memcpy(&buf, src, sizeof(buf));
-	if (IS_INT(buf))
+	if (INT_IS_INT(buf))
 		return tds_convert_int((TDS_INT) buf, desttype, cr);
 
 	switch (desttype) {
@@ -790,8 +807,8 @@ tds_convert_uint8(const TDS_UINT8 *src, int desttype, CONV_RESULT * cr)
 	TDS_CHAR tmp_str[24];
 
 	memcpy(&buf, src, sizeof(buf));
-	/* IS_INT does not work here due to unsigned/signed conversions */
-	if (buf <= (TDS_UINT8) TDS_INT_MAX)
+	/* INT_IS_INT does not work here due to unsigned/signed conversions */
+	if (buf <= TDS_INT_MAX)
 		return tds_convert_int((TDS_INT) buf, desttype, cr);
 
 	switch (desttype) {
@@ -815,7 +832,7 @@ tds_convert_uint8(const TDS_UINT8 *src, int desttype, CONV_RESULT * cr)
 		return sizeof(TDS_UINT);
 		break;
 	case SYBINT8:
-		if (buf > (TDS_UINT8) TDS_INT8_MAX)
+		if (buf > TDS_INT8_MAX)
 			return TDS_CONVERT_OVERFLOW;
 		cr->bi = (TDS_INT8) buf;
 		return sizeof(TDS_INT8);
@@ -894,7 +911,7 @@ tds_convert_numeric(const TDS_NUMERIC * src, int desttype, CONV_RESULT * cr)
 		i = TDS_GET_UA2BE(&(cr->n.array[2]));
 		if (cr->n.array[0])
 			i = -i;
-		if (((i >> 15) ^ cr->n.array[0]) & 1)
+		if (i != 0 && ((i >> 15) ^ cr->n.array[0]) & 1)
 			return TDS_CONVERT_OVERFLOW;
 		cr->si = (TDS_SMALLINT) i;
 		return sizeof(TDS_SMALLINT);
@@ -904,9 +921,9 @@ tds_convert_numeric(const TDS_NUMERIC * src, int desttype, CONV_RESULT * cr)
 		ret = tds_numeric_change_prec_scale(&(cr->n), 5, 0);
 		if (ret < 0)
 			return ret;
-		if (cr->n.array[0] || cr->n.array[1])
-			return TDS_CONVERT_OVERFLOW;
 		i = TDS_GET_UA2BE(&(cr->n.array[2]));
+		if ((i != 0 && cr->n.array[0]) || cr->n.array[1])
+			return TDS_CONVERT_OVERFLOW;
 		cr->usi = (TDS_USMALLINT) i;
 		return sizeof(TDS_USMALLINT);
 		break;
@@ -920,7 +937,7 @@ tds_convert_numeric(const TDS_NUMERIC * src, int desttype, CONV_RESULT * cr)
 		i = TDS_GET_UA4BE(&(cr->n.array[2]));
 		if (cr->n.array[0])
 			i = -i;
-		if (((i >> 31) ^ cr->n.array[0]) & 1)
+		if (i != 0 && ((i >> 31) ^ cr->n.array[0]) & 1)
 			return TDS_CONVERT_OVERFLOW;
 		cr->i = i;
 		return sizeof(TDS_INT);
@@ -930,9 +947,9 @@ tds_convert_numeric(const TDS_NUMERIC * src, int desttype, CONV_RESULT * cr)
 		ret = tds_numeric_change_prec_scale(&(cr->n), 10, 0);
 		if (ret < 0)
 			return ret;
-		if (cr->n.array[0] || cr->n.array[1])
-			return TDS_CONVERT_OVERFLOW;
 		i = TDS_GET_UA4BE(&(cr->n.array[2]));
+		if ((i != 0 && cr->n.array[0]) || cr->n.array[1])
+			return TDS_CONVERT_OVERFLOW;
 		cr->ui = i;
 		return sizeof(TDS_UINT);
 		break;
@@ -947,7 +964,7 @@ tds_convert_numeric(const TDS_NUMERIC * src, int desttype, CONV_RESULT * cr)
 		bi = (bi << 32) + TDS_GET_UA4BE(&(cr->n.array[6]));
 		if (cr->n.array[0])
 			bi = -bi;
-		if (((bi >> 63) ^ cr->n.array[0]) & 1)
+		if (bi != 0 && ((bi >> 63) ^ cr->n.array[0]) & 1)
 			return TDS_CONVERT_OVERFLOW;
 		cr->bi = bi;
 		return sizeof(TDS_INT8);
@@ -957,10 +974,10 @@ tds_convert_numeric(const TDS_NUMERIC * src, int desttype, CONV_RESULT * cr)
 		ret = tds_numeric_change_prec_scale(&(cr->n), 20, 0);
 		if (ret < 0)
 			return ret;
-		if (cr->n.array[0] || cr->n.array[1])
-			return TDS_CONVERT_OVERFLOW;
 		bi = TDS_GET_UA4BE(&(cr->n.array[2]));
 		bi = (bi << 32) + TDS_GET_UA4BE(&(cr->n.array[6]));
+		if ((bi != 0 && cr->n.array[0]) || cr->n.array[1])
+			return TDS_CONVERT_OVERFLOW;
 		cr->ubi = bi;
 		return sizeof(TDS_UINT8);
 		break;
@@ -984,7 +1001,7 @@ tds_convert_numeric(const TDS_NUMERIC * src, int desttype, CONV_RESULT * cr)
 		i = TDS_GET_UA4BE(&(cr->n.array[2]));
 		if (cr->n.array[0])
 			i = -i;
-		if (((i >> 31) ^ cr->n.array[0]) & 1)
+		if (i != 0 && ((i >> 31) ^ cr->n.array[0]) & 1)
 			return TDS_CONVERT_OVERFLOW;
 		cr->m4.mny4 = i;
 		return sizeof(TDS_MONEY4);
@@ -1000,7 +1017,7 @@ tds_convert_numeric(const TDS_NUMERIC * src, int desttype, CONV_RESULT * cr)
 		bi = (bi << 32) + TDS_GET_UA4BE(&(cr->n.array[6]));
 		if (cr->n.array[0])
 			bi = -bi;
-		if (((bi >> 63) ^ cr->n.array[0]) & 1)
+		if (bi != 0 && ((bi >> 63) ^ cr->n.array[0]) & 1)
 			return TDS_CONVERT_OVERFLOW;
 		cr->m.mny = bi;
 		return sizeof(TDS_MONEY);
@@ -1113,7 +1130,7 @@ tds_convert_money4(const TDSCONTEXT * tds_ctx, const TDS_MONEY4 * src, int destt
 		break;
 	case SYBUINT8:
 		dollars = mny.mny4 / 10000;
-		if (!IS_UINT8(dollars))
+		if (dollars < 0)
 			return TDS_CONVERT_OVERFLOW;
 		cr->ubi = dollars;
 		return sizeof(TDS_UINT8);
@@ -1195,7 +1212,7 @@ tds_convert_money(const TDSCONTEXT * tds_ctx, const TDS_MONEY * src, int desttyp
 		break;
 	case SYBINT4:
 		dollars = mymoney / 10000;
-		if (!IS_INT(dollars))
+		if (!INT_IS_INT(dollars))
 			return TDS_CONVERT_OVERFLOW;
 		cr->i = (TDS_INT) dollars;
 		return sizeof(TDS_INT);
@@ -1232,7 +1249,7 @@ tds_convert_money(const TDSCONTEXT * tds_ctx, const TDS_MONEY * src, int desttyp
 		return sizeof(TDS_REAL);
 		break;
 	case SYBMONEY4:
-		if (!IS_INT(mymoney))
+		if (!INT_IS_INT(mymoney))
 			return TDS_CONVERT_OVERFLOW;
 		cr->m4.mny4 = (TDS_INT) mymoney;
 		return sizeof(TDS_MONEY4);
@@ -1262,20 +1279,24 @@ static TDS_INT
 tds_convert_datetimeall(const TDSCONTEXT * tds_ctx, int srctype, const TDS_DATETIMEALL * dta, int desttype, CONV_RESULT * cr)
 {
 	char whole_date_string[64];
+	const char *datetime_fmt;
 	TDSDATEREC when;
 
 	switch (desttype) {
 	case TDS_CONVERT_CHAR:
 	case CASE_ALL_CHAR:
 		tds_datecrack(srctype, dta, &when);
-		tds_strftime(whole_date_string, sizeof(whole_date_string), tds_ctx->locale->date_fmt, &when, 
+		datetime_fmt = tds_ctx->locale->datetime_fmt;
+		if (srctype == SYBMSDATE && tds_ctx->locale->date_fmt)
+			datetime_fmt = tds_ctx->locale->date_fmt;
+		if (srctype == SYBMSTIME && tds_ctx->locale->time_fmt)
+			datetime_fmt = tds_ctx->locale->time_fmt;
+		tds_strftime(whole_date_string, sizeof(whole_date_string), datetime_fmt, &when,
 		             dta->time_prec);
 
 		return string_to_result(desttype, whole_date_string, cr);
 	case SYBDATETIME:
-		if (!IS_INT(dta->date))
-			return TDS_CONVERT_OVERFLOW;
-		cr->dt.dtdays = (TDS_INT) dta->date;
+		cr->dt.dtdays = dta->date;
 		cr->dt.dttime = (TDS_INT) ((dta->time * 3u + 50000u) / 100000u);
 		return sizeof(TDS_DATETIME);
 	case SYBDATETIME4:
@@ -1291,9 +1312,7 @@ tds_convert_datetimeall(const TDSCONTEXT * tds_ctx, int srctype, const TDS_DATET
 		cr->dta = *dta;
 		return sizeof(TDS_DATETIMEALL);
 	case SYBDATE:
-		if (!IS_INT(dta->date))
-			return TDS_CONVERT_OVERFLOW;
-		cr->date = (TDS_INT) dta->date;
+		cr->date = dta->date;
 		return sizeof(TDS_DATE);
 	case SYBTIME:
 		cr->time = (TDS_INT) ((dta->time * 3u + 50000u) / 100000u);
@@ -1337,7 +1356,7 @@ tds_convert_datetime(const TDSCONTEXT * tds_ctx, const TDS_DATETIME * dt, int de
 	case TDS_CONVERT_CHAR:
 	case CASE_ALL_CHAR:
 		tds_datecrack(SYBDATETIME, dt, &when);
-		tds_strftime(whole_date_string, sizeof(whole_date_string), tds_ctx->locale->date_fmt, &when, 3);
+		tds_strftime(whole_date_string, sizeof(whole_date_string), tds_ctx->locale->datetime_fmt, &when, 3);
 
 		return string_to_result(desttype, whole_date_string, cr);
 	case SYBDATETIME:
@@ -1421,33 +1440,37 @@ tds_convert_datetime4(const TDSCONTEXT * tds_ctx, const TDS_DATETIME4 * dt4, int
 static TDS_INT
 tds_convert_time(const TDSCONTEXT * tds_ctx, const TDS_TIME * time, int desttype, CONV_RESULT * cr)
 {
-	TDS_DATETIME dt;
+	TDS_DATETIMEALL dta;
 
 	if (desttype == SYBTIME) {
 		cr->time = *time;
 		return sizeof(TDS_TIME);
 	}
 
-	/* convert to DATETIME and use tds_convert_datetime */
-	dt.dtdays = 0;
-	dt.dttime = *time;
-	return tds_convert_datetime(tds_ctx, &dt, desttype, 0, cr);
+	/* convert to DATETIMEALL and use tds_convert_datetimeall */
+	memset(&dta, 0, sizeof(dta));
+	dta.has_time = 1;
+	dta.time_prec = 3;
+	dta.time = ((TDS_UINT8) *time * 100000 + 2) / 3;
+	dta.time = (((TDS_UINT8) *time) * 20 + 3) / 6 * 10000;
+	return tds_convert_datetimeall(tds_ctx, SYBMSTIME, &dta, desttype, cr);
 }
 
 static TDS_INT
 tds_convert_date(const TDSCONTEXT * tds_ctx, const TDS_DATE * date, int desttype, CONV_RESULT * cr)
 {
-	TDS_DATETIME dt;
+	TDS_DATETIMEALL dta;
 
 	if (desttype == SYBDATE) {
 		cr->date = *date;
 		return sizeof(TDS_DATE);
 	}
 
-	/* convert to DATETIME and use tds_convert_datetime */
-	dt.dtdays = *date;
-	dt.dttime = 0;
-	return tds_convert_datetime(tds_ctx, &dt, desttype, 0, cr);
+	/* convert to DATETIMEALL and use tds_convert_datetimeall */
+	memset(&dta, 0, sizeof(dta));
+	dta.has_date = 1;
+	dta.date = *date;
+	return tds_convert_datetimeall(tds_ctx, SYBMSDATE, &dta, desttype, cr);
 }
 
 static TDS_INT
@@ -1497,7 +1520,7 @@ tds_convert_real(const TDS_REAL* src, int desttype, CONV_RESULT * cr)
 {
 	TDS_REAL the_value;
 
-/* FIXME how many big should be this buffer ?? */
+	/* FIXME how big should be this buffer ?? */
 	char tmp_str[128];
 	TDS_INT mymoney4;
 	TDS_INT8 mymoney;
@@ -1530,7 +1553,7 @@ tds_convert_real(const TDS_REAL* src, int desttype, CONV_RESULT * cr)
 		return sizeof(TDS_USMALLINT);
 		break;
 	case SYBINT4:
-		if (!IS_INT(the_value))
+		if (!FLOAT_IS_INT(the_value))
 			return TDS_CONVERT_OVERFLOW;
 		cr->i = (TDS_INT) the_value;
 		return sizeof(TDS_INT);
@@ -1542,13 +1565,13 @@ tds_convert_real(const TDS_REAL* src, int desttype, CONV_RESULT * cr)
 		return sizeof(TDS_UINT);
 		break;
 	case SYBINT8:
-		if (the_value > (TDS_REAL) TDS_INT8_MAX || the_value < (TDS_REAL) TDS_INT8_MIN)
+		if (!FLOAT_IS_INT8(the_value))
 			return TDS_CONVERT_OVERFLOW;
 		cr->bi = (TDS_INT8) the_value;
 		return sizeof(TDS_INT8);
 		break;
 	case SYBUINT8:
-		if (the_value > (TDS_REAL) TDS_UINT8_MAX || the_value < 0)
+		if (!FLOAT_IS_UINT8(the_value))
 			return TDS_CONVERT_OVERFLOW;
 		cr->ubi = (TDS_UINT8) the_value;
 		return sizeof(TDS_UINT8);
@@ -1639,7 +1662,7 @@ tds_convert_flt8(const TDS_FLOAT* src, int desttype, CONV_RESULT * cr)
 		return sizeof(TDS_USMALLINT);
 		break;
 	case SYBINT4:
-		if (!IS_INT(the_value))
+		if (!FLOAT_IS_INT(the_value))
 			return TDS_CONVERT_OVERFLOW;
 		cr->i = (TDS_INT) the_value;
 		return sizeof(TDS_INT);
@@ -1651,13 +1674,13 @@ tds_convert_flt8(const TDS_FLOAT* src, int desttype, CONV_RESULT * cr)
 		return sizeof(TDS_UINT);
 		break;
 	case SYBINT8:
-		if (the_value > (TDS_FLOAT) TDS_INT8_MAX || the_value < (TDS_FLOAT) TDS_INT8_MIN)
+		if (!FLOAT_IS_INT8(the_value))
 			return TDS_CONVERT_OVERFLOW;
 		cr->bi = (TDS_INT8) the_value;
 		return sizeof(TDS_INT8);
 		break;
 	case SYBUINT8:
-		if (the_value > (TDS_FLOAT) TDS_UINT8_MAX || the_value < 0)
+		if (!FLOAT_IS_UINT8(the_value))
 			return TDS_CONVERT_OVERFLOW;
 		cr->ubi = (TDS_UINT8) the_value;
 		return sizeof(TDS_UINT8);
@@ -1845,11 +1868,17 @@ tds_convert_to_binary(int srctype, const TDS_CHAR * src, TDS_UINT srclen, int de
 
 		ib = cr->cb.ib;
 		if (desttype != TDS_CONVERT_BINARY) {
+			TDS_INT destlen;
+
 			cr->ib = tds_new(TDS_CHAR, (srclen + 2u) / 2u);
 			test_alloc(cr->ib);
 			ib = cr->ib;
+			destlen = tds_char2hex(ib, 0xffffffffu, src, srclen);
+			if (destlen < 0)
+				TDS_ZERO_FREE(cr->ib);
+			return destlen;
 		}
-		return tds_char2hex(ib, desttype == TDS_CONVERT_BINARY ? cr->cb.len : 0xffffffffu, src, srclen);
+		return tds_char2hex(ib, cr->cb.len, src, srclen);
 
 	default:
 		return TDS_CONVERT_NOAVAIL;
@@ -1875,14 +1904,14 @@ tds_convert_to_binary(int srctype, const TDS_CHAR * src, TDS_UINT srclen, int de
  * @return length of result or TDS_CONVERT_* failure code on failure. All TDS_CONVERT_* constants are <0.
  */
 TDS_INT
-tds_convert(const TDSCONTEXT * tds_ctx, int srctype, const TDS_CHAR * src, TDS_UINT srclen, int desttype, CONV_RESULT * cr)
+tds_convert(const TDSCONTEXT *tds_ctx, int srctype, const void *src, TDS_UINT srclen, int desttype, CONV_RESULT *cr)
 {
 	TDS_INT length = 0;
 
 	assert(srclen >= 0 && srclen <= 2147483647u);
 
 	if (srctype == SYBVARIANT) {
-		TDSVARIANT *v = (TDSVARIANT*) src;
+		const TDSVARIANT *v = (const TDSVARIANT*) src;
 		srctype = v->type;
 		src = v->data;
 		srclen = v->data_len;
@@ -1971,6 +2000,7 @@ tds_convert(const TDSCONTEXT * tds_ctx, int srctype, const TDS_CHAR * src, TDS_U
 		break;
 	case SYBNVARCHAR:
 	case SYBNTEXT:
+	case SYBMSTABLE:
 	default:
 		return TDS_CONVERT_NOAVAIL;
 		break;
@@ -1988,11 +2018,10 @@ tds_convert(const TDSCONTEXT * tds_ctx, int srctype, const TDS_CHAR * src, TDS_U
 static int
 string_to_datetime(const char *instr, TDS_UINT len, int desttype, CONV_RESULT * cr)
 {
-	enum states
-	{ GOING_IN_BLIND,
+	enum states {
+		GOING_IN_BLIND,
 		PUT_NUMERIC_IN_CONTEXT,
-		DOING_ALPHABETIC_DATE,
-		STRING_GARBLED
+		DOING_ALPHABETIC_DATE
 	};
 
 	char *in;
@@ -2009,7 +2038,7 @@ string_to_datetime(const char *instr, TDS_UINT len, int desttype, CONV_RESULT *
 	TDS_INT dt_days;
 	int i;
 
-	int current_state;
+	enum states current_state;
 
 	memset(&t, '\0', sizeof(t));
 	t.tm_mday = 1;
@@ -2036,7 +2065,7 @@ string_to_datetime(const char *instr, TDS_UINT len, int desttype, CONV_RESULT *
 					monthdone++;
 					current_state = DOING_ALPHABETIC_DATE;
 				} else {
-					current_state = STRING_GARBLED;
+					goto string_garbled;
 				}
 			}
 
@@ -2072,13 +2101,13 @@ string_to_datetime(const char *instr, TDS_UINT len, int desttype, CONV_RESULT *
 					if (store_yymmdd_date(tok, &t))
 						current_state = GOING_IN_BLIND;
 					else
-						current_state = STRING_GARBLED;
+						goto string_garbled;
 					break;
 
 					/* anything else is nonsense...               */
 
 				default:
-					current_state = STRING_GARBLED;
+					goto string_garbled;
 					break;
 				}
 			}
@@ -2099,7 +2128,7 @@ string_to_datetime(const char *instr, TDS_UINT len, int desttype, CONV_RESULT *
 				current_state = GOING_IN_BLIND;
 			} else {
 				tdsdump_log(TDS_DBG_INFO1, "string_to_datetime: string_garbled\n");
-				current_state = STRING_GARBLED;
+				goto string_garbled;
 			}
 
 			break;	/* end of GOING_IN_BLIND */
@@ -2114,11 +2143,11 @@ string_to_datetime(const char *instr, TDS_UINT len, int desttype, CONV_RESULT *
 					else
 						current_state = DOING_ALPHABETIC_DATE;
 				} else {
-					current_state = STRING_GARBLED;
+					goto string_garbled;
 				}
 			} else if (is_numeric(tok)) {
 				if (mdaydone && yeardone)
-					current_state = STRING_GARBLED;
+					goto string_garbled;
 				else
 					switch (strlen(tok)) {
 					case 4:
@@ -2151,10 +2180,10 @@ string_to_datetime(const char *instr, TDS_UINT len, int desttype, CONV_RESULT *
 						break;
 
 					default:
-						current_state = STRING_GARBLED;
+						goto string_garbled;
 					}
 			} else {
-				current_state = STRING_GARBLED;
+				goto string_garbled;
 			}
 
 			break;	/* end of DOING_ALPHABETIC_DATE */
@@ -2174,7 +2203,7 @@ string_to_datetime(const char *instr, TDS_UINT len, int desttype, CONV_RESULT *
 					store_hour(last_token, tok, &t);
 					current_state = GOING_IN_BLIND;
 				} else {
-					current_state = STRING_GARBLED;
+					goto string_garbled;
 				}
 			} else if (is_numeric(tok)) {
 				switch (strlen(tok)) {
@@ -2192,20 +2221,13 @@ string_to_datetime(const char *instr, TDS_UINT len, int desttype, CONV_RESULT *
 					break;
 
 				default:
-					current_state = STRING_GARBLED;
+					goto string_garbled;
 				}
 			} else {
-				current_state = STRING_GARBLED;
+				goto string_garbled;
 			}
 
 			break;	/* end of PUT_NUMERIC_IN_CONTEXT */
-
-		case STRING_GARBLED:
-
-			tdsdump_log(TDS_DBG_INFO1,
-				    "error_handler:  Attempt to convert data stopped by syntax error in source field \n");
-			free(in);
-			return TDS_CONVERT_SYNTAX;
 		}
 
 		tok = strtok_r(NULL, " ,", &lasts);
@@ -2256,6 +2278,12 @@ string_to_datetime(const char *instr, TDS_UINT len, int desttype, CONV_RESULT *
 	cr->dta.time_prec = 7; /* TODO correct value */
 	cr->dta.time = ((TDS_UINT8) dt_time) * 10000000u + t.tm_ns / 100u;
 	return sizeof(TDS_DATETIMEALL);
+
+string_garbled:
+	tdsdump_log(TDS_DBG_INFO1,
+		    "error_handler:  Attempt to convert data stopped by syntax error in source field \n");
+	free(in);
+	return TDS_CONVERT_SYNTAX;
 }
 
 static int
@@ -2957,6 +2985,17 @@ tds_get_null_type(TDS_SERVER_TYPE srctype)
 	return srctype;
 }
 
+static void
+two_digit(char *out, int num)
+{
+	if (num < 1)
+		num = 1;
+	if (num > 31)
+		num = 31;
+	out[0] = num < 10 ? ' ' : num/10 + '0';
+	out[1] = num%10 + '0';
+}
+
 /**
  * format a date string according to an "extended" strftime(3) formatting definition.
  * @param buf     output buffer
@@ -2973,7 +3012,8 @@ tds_strftime(char *buf, size_t maxsize, const char *format, const TDSDATEREC * d
 
 	size_t length;
 	char *our_format;
-	char *pz = NULL;
+	char *pz;
+	bool z_found = false;
 	
 	assert(buf);
 	assert(format);
@@ -2998,33 +3038,60 @@ tds_strftime(char *buf, size_t maxsize, const char *format, const TDSDATEREC * d
 #endif
 
 	/* more characters are required because we replace %z with up to 7 digits */
-	our_format = tds_new(char, strlen(format) + 1 + 5);
+	our_format = tds_new(char, strlen(format) + 1 + 5 + 1);
 	if (!our_format)
 		return 0;
 
 	strcpy(our_format, format);
 
-	/*
-	 * Look for "%z" in the format string.  If found, replace it with dr->milliseconds.
-	 * For example, if milliseconds is 124, the format string
-	 * "%b %d %Y %H:%M:%S.%z" would become
-	 * "%b %d %Y %H:%M:%S.124".
-	 */
-	for (pz = our_format; (pz = strstr(pz, "%z")) != NULL; pz++) {
-		/* Skip any escaped cases (%%z) */
-		if (pz > our_format && *(pz - 1) != '%')
-			break;
-	}
+	for (pz = our_format; *pz; ) {
+		if (*pz++ != '%')
+			continue;
 
-	if (pz) {
-		if (prec || pz <= our_format || pz[-1] != '.') {
-			char buf[12];
-			sprintf(buf, "%07d", dr->decimicrosecond);
-			memcpy(pz, buf, prec);
-			strcpy(pz + prec, format + (pz - our_format) + 2);
-		} else {
-			strcpy(pz - 1, format + (pz - our_format) + 2);
+		switch (*pz) {
+		case 0:
+			*pz++ = '%';
+			*pz = 0;
+			continue;
+		case '%':
+			/* escaped, do not treat as format on next iteration */
+			break;
+		case 'e':
+			/* not portable: day of month, single digit preceded by a blank */
+			/* not supported on old Windows versions */
+			two_digit(pz-1, dr->day);
+			break;
+		case 'l':
+			/* not portable: 12-hour, single digit preceded by a blank */
+			/* not supported on: SCO Unix, AIX, HP-UX, Windows
+			 * supported on: *BSD, MacOS, Linux, Solaris */
+			two_digit(pz-1, (dr->hour + 11u) % 12u + 1);
+			break;
+		case 'z':
+			/*
+			 * Look for "%z" in the format string.  If found, replace it with dr->milliseconds.
+			 * For example, if milliseconds is 124, the format string
+			 * "%b %d %Y %H:%M:%S.%z" would become
+			 * "%b %d %Y %H:%M:%S.124".
+			 */
+			if (z_found)
+				break;
+			z_found = true;
+
+			--pz;
+			if (prec || pz <= our_format || pz[-1] != '.') {
+				char buf[12];
+				sprintf(buf, "%07d", dr->decimicrosecond);
+				memcpy(pz, buf, prec);
+				strcpy(pz + prec, format + (pz - our_format) + 2);
+				pz += prec;
+			} else {
+				strcpy(pz - 1, format + (pz - our_format) + 2);
+				pz--;
+			}
+			continue;
 		}
+		++pz;
 	}
 
 	length = strftime(buf, maxsize, our_format, &tm);
